import asyncio

import rerun as rr
from loguru import logger
from nova import MotionGroup
from scipy.spatial.transform import Rotation as R

from nova_rerun_bridge import colors
from nova_rerun_bridge.consts import TIME_INTERVAL_NAME
from nova_rerun_bridge.dh_robot import DHRobot
from nova_rerun_bridge.robot_visualizer import RobotVisualizer


def log_joint_positions_once(motion_group: str, robot: DHRobot, joint_position):
    """Compute and log joint positions for a robot."""
    joint_positions = robot.calculate_joint_positions(joint_position)
    line_segments = [
        [joint_positions[i], joint_positions[i + 1]] for i in range(len(joint_positions) - 1)
    ]
    segment_colors = [colors.colors[i % len(colors.colors)] for i in range(len(line_segments))]

    rr.log(
        f"{motion_group}/dh_parameters",
        rr.LineStrips3D(line_segments, colors=segment_colors),
        static=True,
    )


class MotionGroupProcessor:
    def __init__(self):
        self.last_tcp_pose = {}

    def tcp_pose_changed(self, motion_group: str, tcp_pose) -> bool:
        """Check if the TCP pose has changed compared to the last logged value."""
        last_pose = self.last_tcp_pose.get(motion_group)
        current_position = [tcp_pose.position.x, tcp_pose.position.y, tcp_pose.position.z]
        current_rotation = [tcp_pose.orientation.x, tcp_pose.orientation.y, tcp_pose.orientation.z]

        if last_pose is None:
            # No previous pose, consider it as changed
            self.last_tcp_pose[motion_group] = (current_position, current_rotation)
            return True

        last_position, last_rotation = last_pose
        if current_position != last_position or current_rotation != last_rotation:
            # Update the cache and return True if either position or rotation has changed
            self.last_tcp_pose[motion_group] = (current_position, current_rotation)
            return True

        return False

    def log_tcp_orientation(self, motion_group: str, tcp_pose):
        """Log TCP orientation and position."""
        rotation_vector = [tcp_pose.orientation.x, tcp_pose.orientation.y, tcp_pose.orientation.z]
        rotation = R.from_rotvec(rotation_vector)
        angle = rotation.magnitude() if rotation.magnitude() != 0 else 0.0
        axis_angle = rotation.as_rotvec() / angle if angle != 0 else [0, 0, 0]

        rr.log(
            f"{motion_group}/tcp_position",
            rr.Transform3D(
                translation=[tcp_pose.position.x, tcp_pose.position.y, tcp_pose.position.z],
                rotation=rr.RotationAxisAngle(axis=axis_angle, angle=angle),
            ),
            static=True,
        )


async def stream_motion_group(self, motion_group: MotionGroup) -> None:
    """Stream individual motion group state to Rerun."""

    rr.set_time_seconds(TIME_INTERVAL_NAME, 0)

    processor = MotionGroupProcessor()

    motion_groups = await self.nova._api_client.motion_group_api.list_motion_groups(
        self.nova.cell()._cell_id
    )
    motion_motion_group = next(
        (mg for mg in motion_groups.instances if mg.motion_group == motion_group.motion_group_id),
        None,
    )

    try:
        optimizer_config = (
            await self.nova._api_client.motion_group_infos_api.get_optimizer_configuration(
                self.nova.cell()._cell_id, motion_group.motion_group_id
            )
        )

        robot = DHRobot(optimizer_config.dh_parameters, optimizer_config.mounting)
        visualizer = RobotVisualizer(
            robot=robot,
            robot_model_geometries=optimizer_config.safety_setup.robot_model_geometries,
            tcp_geometries=optimizer_config.safety_setup.tcp_geometries,
            static_transform=True,
            base_entity_path=motion_group.motion_group_id,
            albedo_factor=[0, 255, 100],
            model_from_controller=motion_motion_group.model_from_controller,
        )

        logger.info(f"Started streaming motion group {motion_group}")
        async for state in self.nova._api_client.motion_group_infos_api.stream_motion_group_state(
            self.nova.cell()._cell_id, motion_group.motion_group_id
        ):
            if processor.tcp_pose_changed(motion_group, state.state.tcp_pose):
                # Log joint positions
                log_joint_positions_once(
                    motion_group.motion_group_id, robot, state.state.joint_position
                )

                # Log robot geometries
                visualizer.log_robot_geometry(state.state.joint_position)

                processor.log_tcp_orientation(motion_group.motion_group_id, state.state.tcp_pose)

        await asyncio.sleep(0.01)  # Prevents CPU overuse
    except asyncio.CancelledError:
        logger.info(f"Stopped streaming motion group {motion_group}")
    except Exception as e:
        logger.error(f"Error streaming motion group {motion_group}: {e}")



import asyncio

from nova import MotionSettings
from nova.actions import jnt, ptp
from nova.api import models
from nova.core.exceptions import PlanTrajectoryFailed
from nova.core.nova import Nova
from nova.types import Pose

from nova_rerun_bridge import NovaRerunBridge


async def test():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()

        cell = nova.cell()
        controller = await cell.ensure_virtual_robot_controller(
            "ur",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR10E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        # Connect to the controller and activate motion groups
        async with controller[0] as motion_group:
            home_joints = await motion_group.joints()
            tcp_names = await motion_group.tcp_names()
            tcp = tcp_names[0]

            # Get current TCP pose and offset it slightly along the x-axis
            current_pose = await motion_group.tcp_pose(tcp)
            target_pose = current_pose @ Pose((1, 0, 0, 0, 0, 0))

            actions = [
                jnt(home_joints),
                ptp(target_pose @ [-100, 0, 0, 0, 0, 0]),
                ptp(target_pose @ [-200, 0, 0, 0, 0, 0]),
                ptp(target_pose @ [-500, 0, 0, 0, 0, 0]),
                ptp(target_pose @ [-2000, 0, 0, 0, 0, 0]),
                jnt(home_joints),
            ]

            # you can update the settings of the action
            for action in actions:
                action.settings = MotionSettings(tcp_velocity_limit=200)

            try:
                joint_trajectory = await motion_group.plan(actions, tcp)
                await bridge.log_actions(actions)
                await bridge.log_trajectory(joint_trajectory, tcp, motion_group)
                await motion_group.execute(joint_trajectory, tcp, actions=actions)
            except PlanTrajectoryFailed as e:
                await bridge.log_actions(actions, show_connection=True)
                await bridge.log_trajectory(e.error.joint_trajectory, tcp, motion_group)
                await bridge.log_error_feedback(e.error.error_feedback)
                return

        await cell.delete_robot_controller(controller.name)


if __name__ == "__main__":
    asyncio.run(test())



import asyncio

from nova import Controller, Nova
from nova.actions import MotionSettings, Pose, jnt, ptp
from numpy import pi

from nova_rerun_bridge import NovaRerunBridge
from nova_rerun_bridge.trajectory import TimingMode

"""
Example: Show an external axis with a robot moving in a cell.

Prerequisites:
- A cell with a robot with an positioner (external) axis: setup a yaskawa-ar1440 and import the yaskawa-ar1440-with-external-axis.yaml in the settings app
- Set mounting of 16@yaskawa to yaskawa-ar1440-16-mounting
"""


async def move_robot(controller: Controller, bridge: NovaRerunBridge):
    async with controller[0] as motion_group:
        tcp_names = await motion_group.tcp_names()
        tcp = tcp_names[0]
        home_joints = await motion_group.joints()

        current_pose = await motion_group.tcp_pose(tcp)

        actions = [
            jnt(home_joints),
            ptp(current_pose @ Pose((0, 0, -100, 0, -pi / 2, 0))),
            ptp(current_pose @ Pose((-500, 0, 0, 0, -pi / 2, 0))),
            jnt(home_joints),
        ]

        for action in actions:
            action.settings = MotionSettings(tcp_velocity_limit=200)

        trajectory = await motion_group.plan(actions, tcp)
        await bridge.log_trajectory(trajectory, tcp, motion_group, timing_mode=TimingMode.SYNC)


async def move_positioner(controller: Controller, bridge: NovaRerunBridge):
    async with controller[16] as motion_group:
        actions = [jnt([0, 0]), jnt([pi / 4, pi / 4]), jnt([-pi / 4, -pi / 4]), jnt([0, 0])]

        trajectory = await motion_group.plan(actions, "")
        await bridge.log_trajectory(trajectory, "", motion_group, timing_mode=TimingMode.SYNC)


async def main():
    nova = Nova()
    bridge = NovaRerunBridge(nova)
    await bridge.setup_blueprint()
    cell = nova.cell()
    yaskawa = await cell.controller("yaskawa")
    await asyncio.gather(
        move_robot(yaskawa, bridge=bridge), move_positioner(yaskawa, bridge=bridge)
    )
    await nova.close()
    await bridge.cleanup()


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import signal
from contextlib import asynccontextmanager

from nova.api import models
from nova.core.nova import Nova

from nova_rerun_bridge import NovaRerunBridge


@asynccontextmanager
async def handle_shutdown():
    stop = asyncio.Future()

    def signal_handler():
        if not stop.done():
            stop.set_result(None)

    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, signal_handler)
    try:
        yield stop
    finally:
        loop.remove_signal_handler(signal.SIGINT)


async def test():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()

        cell = nova.cell()
        controller = await cell.ensure_virtual_robot_controller(
            "ur",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR10E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        # Connect to the controller and activate motion groups
        async with controller[0] as motion_group:
            await bridge.start_streaming(motion_group)

            # Keep streaming until Ctrl+C
            async with handle_shutdown() as stop:
                try:
                    await stop
                except asyncio.CancelledError:
                    pass
                finally:
                    await bridge.stop_streaming()


if __name__ == "__main__":
    asyncio.run(test())



import asyncio

from nova import Controller, Nova
from nova.actions import jnt, ptp
from nova.api import models
from wandelbots_api_client.models import (
    CoordinateSystem,
    RotationAngles,
    RotationAngleTypes,
    Vector3d,
)

from nova_rerun_bridge import NovaRerunBridge
from nova_rerun_bridge.trajectory import TimingMode

"""
Example: Move multiple robots simultaneously.

Prerequisites:
- A cell with two robots: one named "ur" and another named "kuka".
"""


async def move_robot(controller: Controller, bridge: NovaRerunBridge):
    async with controller[0] as motion_group:
        home_joints = await motion_group.joints()
        tcp_names = await motion_group.tcp_names()
        tcp = tcp_names[0]

        current_pose = await motion_group.tcp_pose(tcp)
        target_pose = current_pose @ (100, 0, 0, 0, 0, 0)
        actions = [jnt(home_joints), ptp(target_pose), jnt(home_joints)]

        trajectory = await motion_group.plan(actions, tcp)
        await bridge.log_trajectory(trajectory, tcp, motion_group, timing_mode=TimingMode.SYNC)

        await motion_group.plan_and_execute(actions, tcp)


async def main():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        cell = nova.cell()

        ur10 = await cell.ensure_virtual_robot_controller(
            "ur10",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR10E,
            models.Manufacturer.UNIVERSALROBOTS,
        )
        ur5 = await cell.ensure_virtual_robot_controller(
            "ur5",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR5E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        # NC-1047
        await asyncio.sleep(3)

        await nova._api_client.virtual_robot_setup_api.set_virtual_robot_mounting(
            cell="cell",
            controller="ur5",
            id=0,
            coordinate_system=CoordinateSystem(
                coordinate_system="world",
                name="mounting",
                reference_uid="",
                position=Vector3d(x=500, y=0, z=0),
                rotation=RotationAngles(
                    angles=[0, 0, 0], type=RotationAngleTypes.EULER_ANGLES_EXTRINSIC_XYZ
                ),
            ),
        )

        # NC-1047
        await asyncio.sleep(5)

        await bridge.setup_blueprint()
        await asyncio.gather(move_robot(ur5, bridge=bridge), move_robot(ur10, bridge=bridge))

        await cell.delete_robot_controller(ur5.name)
        await cell.delete_robot_controller(ur10.name)


if __name__ == "__main__":
    asyncio.run(main())



import asyncio

from nova import Nova
from nova.actions import WriteAction, jnt, ptp
from nova.api import models
from nova.types import Pose

from nova_rerun_bridge import NovaRerunBridge

"""
Example: Move the robot and set I/Os on the path.

Prerequisites:
- Create an NOVA instance
- Set env variables (you can specify them in an .env file):
    - NOVA_API=<api>
    - NOVA_ACCESS_TOKEN=<token>
"""


async def main():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()
        cell = nova.cell()
        controller = await cell.ensure_virtual_robot_controller(
            "ur",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR10E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        # Connect to the controller and activate motion groups
        async with controller[0] as motion_group:
            home_joints = await motion_group.joints()
            tcp_names = await motion_group.tcp_names()
            tcp = tcp_names[0]

            # Get current TCP pose and offset it slightly along the x-axis
            current_pose = await motion_group.tcp_pose(tcp)
            target_pose = current_pose @ Pose((100, 0, 0, 0, 0, 0))
            actions = [
                jnt(home_joints),
                WriteAction(key="digital_out[0]", value=False),
                ptp(target_pose),
                jnt(home_joints),
            ]

            # io_value = await controller.read_io("digital_out[0]")
            joint_trajectory = await motion_group.plan(actions, tcp)

            await bridge.log_actions(actions)
            await bridge.log_trajectory(joint_trajectory, tcp, motion_group)
            await motion_group.execute(joint_trajectory, tcp, actions=actions)

        await cell.delete_robot_controller(controller.name)


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
from math import pi

from nova import MotionGroup, Nova
from nova.actions import ptp
from nova.api import models
from nova.types import Pose

from nova_rerun_bridge import NovaRerunBridge
from nova_rerun_bridge.trajectory import TimingMode

"""
Example: Move multiple robots to perform coordinated movements.

Prerequisites:
- Create an NOVA instance
- Set env variables (you can specify them in an .env file):
    - NOVA_API=<api>
    - NOVA_ACCESS_TOKEN=<token>
"""


async def move_robot(
    motion_group: MotionGroup, tcp: str, bridge: NovaRerunBridge, timing_mode: TimingMode
):
    home_pose = Pose((200, 200, 600, 0, pi, 0))
    target_pose = home_pose @ (100, 0, 0, 0, 0, 0)
    actions = [
        ptp(home_pose),
        ptp(target_pose),
        ptp(target_pose @ (0, 0, 100, 0, 0, 0)),
        ptp(target_pose @ (0, 100, 0, 0, 0, 0)),
        ptp(home_pose),
    ]

    trajectory = await motion_group.plan(actions, tcp)
    await bridge.log_trajectory(trajectory, tcp, motion_group, timing_mode=timing_mode)

    await motion_group.plan_and_execute(actions, tcp)


async def main():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()
        cell = nova.cell()
        ur10 = await cell.ensure_virtual_robot_controller(
            "ur10",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR10E,
            models.Manufacturer.UNIVERSALROBOTS,
        )
        ur5 = await cell.ensure_virtual_robot_controller(
            "ur5",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR5E,
            models.Manufacturer.UNIVERSALROBOTS,
        )
        tcp = "Flange"

        flange_state = await ur10[0].get_state(tcp)
        print(flange_state)

        # activate all motion groups
        async with ur10:
            await move_robot(ur10.motion_group(0), tcp, bridge, TimingMode.CONTINUE)

        # activate motion group 0
        async with ur10.motion_group(0) as mg_0:
            await move_robot(mg_0, tcp, bridge, TimingMode.CONTINUE)

        # activate motion group 0
        async with ur10[0] as mg_0:
            await move_robot(mg_0, tcp, bridge, TimingMode.CONTINUE)

        # activate motion group 0 from two different controllers
        async with ur10[0] as ur_0_mg, ur5[0] as kuka_0_mg:
            await asyncio.gather(
                move_robot(ur_0_mg, tcp, bridge, TimingMode.SYNC),
                move_robot(kuka_0_mg, tcp, bridge, TimingMode.SYNC),
            )

        bridge.continue_after_sync()

        # activate motion group 0 from two different controllers
        mg_0 = ur10.motion_group(0)
        mg_1 = ur5.motion_group(0)
        async with mg_0, mg_1:
            await asyncio.gather(
                move_robot(mg_0, tcp, bridge, TimingMode.SYNC),
                move_robot(mg_1, tcp, bridge, TimingMode.SYNC),
            )

        await cell.delete_robot_controller(ur5.name)
        await cell.delete_robot_controller(ur10.name)


if __name__ == "__main__":
    asyncio.run(main())
